/*
 * Takes udp requests and processes them
 */
#include <arpa/inet.h> 
#include <linux/ax25.h> 
#include <linux/if.h>
#include <netinet/in.h> 
#include <sys/socket.h> 
#include <sys/time.h>
#include <sys/types.h> 
#include <ctype.h>
#include <netdb.h>
#include <stdio.h>
#include <syslog.h>
#include <string.h>
#include <unistd.h>

#include "rspfax25.h"
#include "rspfudp.h"
#include "queue.h"
#include "rspfd.h"
#include "rspfif.h"

extern int errno;
extern char *sys_errlist[];
extern struct queue *routerq;
extern struct queue *adj_queue;
extern struct queue *linkq;
extern struct queue *ifqueue;
extern struct queue *rt_queue;
extern struct queue *nodgrpq;
extern const char *StatusName[];
extern const char *version;
extern struct rspf_mib rspf_stats;
extern int debug_mode;


void send_adjs(int s);
void send_links(int s);
void send_routers(int s);
void send_interfaces(int s);
void send_stats(int s);
void send_routes(int s);
void send_nodegroups(int s);

char* in_ntoa(u_long addr);

int init_tcp(int port)
{
	struct sockaddr_in asin;
	struct protoent *ppe;
	int s;
	
	bzero((char*)&asin, sizeof(asin));
	asin.sin_family = AF_INET;
	asin.sin_addr.s_addr = htonl(INADDR_ANY);
	asin.sin_port = htons(port);
	
	if ( (ppe = getprotobyname("tcp")) == 0)
	{
		syslog(LOG_DAEMON | LOG_ERR, "init_tcp(): getprotobyname failed (%m)");
		return -1;
	}
	
	if ( (s= socket(PF_INET, SOCK_STREAM, ppe->p_proto)) < 0)
	{
		syslog(LOG_DAEMON | LOG_ERR, "init_tcp(): socket failed (%m)");
		(void)close(s);
		return -1;
	}
	
	if (bind(s, (struct sockaddr *)&asin, sizeof(asin)) < 0)
	{
		syslog(LOG_DAEMON | LOG_ERR, "init_tcp(): bind failed (%m)");
		return -1;
	}
	listen(s, 3);	
	
	return s;

}

int do_tcp(int skt, char *inbuf)
{
	int i;
	char msg[128];
	
	/* Make everyting lowercase */
	for(i=0; i < strlen(inbuf); i++)
	{
		if (inbuf[i] == '\n' || inbuf[i] == '\r')
			inbuf[i] = '\000';
		inbuf[i] = tolower(inbuf[i]);
	}	
	if (debug_mode)
		printf("do_tcp(): Processing command: '%s'.\n", inbuf);
	if (strncmp(inbuf, "adj",3) == 0 )
	{
		send_adjs(skt);
	} else if (strcmp(inbuf, "quit") == 0)
	{
		write(skt, "Goodbye!\n", 10);
		close(skt);
		return -1;
	} else if (strncmp(inbuf, "group", 5) == 0)
	{
		send_nodegroups(skt);
	} else if (strcmp(inbuf, "?") == 0 || strcmp(inbuf, "help") == 0)
	{
		sprintf(msg, "Valid commands are;\n");
		write(skt, msg, strlen(msg));
		sprintf(msg, "  adjs, groups, help, ifaces, links, routes, routers.\n");
		write(skt, msg, strlen(msg));
	} else if (strncmp(inbuf, "iface", 4) == 0)
	{
		send_interfaces(skt);
	} else if (strncmp(inbuf, "link", 4) == 0)
	{
		send_links(skt);
	} else if (strcmp(inbuf, "routes") == 0)
	{
		send_routes(skt);
	} else if (strcmp(inbuf, "routers") == 0)
	{
		send_routers(skt);
	} else if (strncmp(inbuf, "stat", 4) == 0)
	{
		send_stats(skt);
	}
	
	write(skt, "> ", 3);
	return 0;
}
	
void send_adjs(int s)
{
	struct rspf_adj *ptr;
	char tbuf[30];
	char outbuf[128];
	time_t elapsed;
	qmark adj_qm;
	

	ptr = (struct rspf_adj*)qmove_first(adj_queue, &adj_qm);

	sprintf(outbuf, "Callsign IP Address       Port seq #  %% hrd Lastheard\n");
	write(s, outbuf, strlen(outbuf)+1);
      	while(ptr != NULL) 
      	{
      		bzero(outbuf, strlen(outbuf));
		if (ptr->rrhtime != 0) 
      		{
			elapsed = time(NULL) -  ptr->rrhtime;
			strftime(tbuf, 30, "%a %H:%M:%S", localtime(&(ptr->rrhtime)));
		} else
			strcpy(tbuf, "Not Router");
		sprintf(outbuf,"%8s %15s %5s %#6x %3d %14s %10s\n", ax25_ntoa(&ptr->dladdr), in_ntoa(ptr->addr), ptr->port, ptr->tx_pkts, ptr->rx_ratio, tbuf, StatusName[ptr->status]);
		write(s, outbuf, strlen(outbuf)+1);		
		
		ptr = (struct rspf_adj*)qmove_next(adj_queue,&adj_qm);
	}
} /* print_adjs */	

void send_links(int s)
{
	struct link *ptr;
	char buf[200];
	qmark link_qm;
	
	ptr = (struct link*)qmove_first(linkq, &link_qm);
	sprintf(buf, "Link States\nSource Address   Destinatn Address\tCost Horizon Seq #\n");
	write(s, buf, strlen(buf)+1);
	while (ptr != NULL)
	{
		sprintf(buf, "%15s->", in_ntoa(ptr->saddr));
		sprintf(buf+strlen(buf), "%15s/%2d\t%4d (%4d)  %5u\n",
			in_ntoa(ptr->daddr), ptr->sigbits, ptr->cost, ptr->horizon, ptr->seq_no);
		write(s, buf, strlen(buf)+1);		
		ptr = (struct link*)qmove_next(linkq, &link_qm);
	}
} /* send_links */

void send_routers(int s)
{
	struct router *rtr;
	qmark rt_qm;
	char buf[200];
	char tbuf[30];
	
	sprintf(buf, "Router List\nIP Address Sequence sub-seq  Bull Time\n");
	write(s, buf, strlen(buf)+1);
	rtr = (struct router*)qmove_first(routerq, &rt_qm);
	while(rtr != NULL)
	{
		strftime(tbuf, 30, "%a %H:%M:%S", localtime(&(rtr->bulltime)));
		sprintf(buf, "%15s %4u %3u %s\n", in_ntoa(rtr->addr), rtr->seq_no, rtr->sub_seq_no, tbuf);
		write(s, buf, strlen(buf)+1);
		rtr = (struct router*)qmove_next(routerq, &rt_qm);
	}
} /*send routers */

void send_interfaces(int s)
{
	char buf[128];
	struct rspf_if *ifptr;
	qmark if_qm;
	int adjcnt;
	struct rspf_adj *aptr;
	qmark adj_qm;
	
	
	sprintf(buf, "Interfaces\n   Name    Cost Nodes\n");
	write(s, buf, strlen(buf)+1);
	ifptr = (struct rspf_if*)qmove_first(ifqueue, &if_qm);
	while(ifptr != NULL)
	{
		adjcnt = 0;
		aptr = (struct rspf_adj*)qmove_first(adj_queue, &adj_qm);
		while(aptr != NULL)
		{
			if (strcmp(aptr->port, ifptr->name) == 0)
				adjcnt++;
			aptr = (struct rspf_adj*)qmove_next(adj_queue, &adj_qm);
		}
		sprintf(buf, "%10s %4d %4d\n", ifptr->name, ifptr->cost, adjcnt);
		write(s, buf, strlen(buf)+1);
		ifptr = (struct rspf_if*)qmove_next(ifqueue, &if_qm);
	}
	
}

void send_stats(int s)
{
	char buf[128];
	
	sprintf(buf, "System Statistics\nIncoming              \t Outgoing\n");
	write(s, buf, strlen(buf)+1);
	
	sprintf(buf, "Messages       %6lu \t %6lu\n", rspf_stats.rspfInMsgs, rspf_stats.rspfOutMsgs);
	write(s, buf, strlen(buf)+1);
	
	sprintf(buf, "Rtr-Rtr Hellos %6lu \t %6lu\n", rspf_stats.rspfInRrhs, rspf_stats.rspfOutRrhs);
	write(s, buf, strlen(buf)+1);
	
	sprintf(buf, "Rtr Envelopes  %6lu \t %6lu\n", rspf_stats.rspfInRouteEnvs, rspf_stats.rspfOutRouteEnvs);
	write(s, buf, strlen(buf)+1);
	
	sprintf(buf, "Header Errors  %6lu\n", rspf_stats.rspfInHdrErrors);
	write(s, buf, strlen(buf)+1);
	
	sprintf(buf, "Unknown Types  %6lu\n", rspf_stats.rspfInUnknownTypes);
	write(s, buf, strlen(buf)+1);
	
	sprintf(buf, "Not RSPF iface %6lu\n", rspf_stats.rspfInNotIfaces);
	write(s, buf, strlen(buf)+1);
	
	sprintf(buf, "RRH Timer     %6lu\tSus Pings     %6lu\n", rspf_stats.rspfRrhTimer,  rspf_stats.rspfSusPings);
	write(s, buf, strlen(buf)+1);
	
	sprintf(buf, "Sus Timeout   %6lu\tPing Timer    %6lu\tBullitn Timer %6lu\n", rspf_stats.rspfSusTimeout, rspf_stats.rspfPingTimer, rspf_stats.rspfBullTimer);
	write(s, buf, strlen(buf)+1);
	
	sprintf(buf, "Bull Timeout  %6lu\n", rspf_stats.rspfBullTimeout);
	write(s, buf, strlen(buf)+1);
	
	sprintf(buf, "Current Adjs  %6lu\tCurrent Iface %6lu\n", rspf_stats.rspfCurrAdjacencies, rspf_stats.rspfCurrIfaces);
	write(s, buf, strlen(buf)+1);

	sprintf(buf, "Envelope No   %6u\tSequence No   %5u\tSub-Seq No    %6u\n", rspf_stats.EnvelopeNumber, rspf_stats.SequenceNumber, rspf_stats.SubSequenceNumber);
	write(s, buf, strlen(buf)+1); 
	
	sprintf(buf, "Link Horizon  %6u\tGroup Horizon %6u\tLocal Horizon %6u\n", rspf_stats.rspfLinkHorizon, rspf_stats.rspfGroupHorizon, rspf_stats.rspfLocalHorizon);
	write(s, buf, strlen(buf)+1);
	
	sprintf(buf, "Port Horizon  %6u\n", rspf_stats.rspfPortableHorizon);
	write(s, buf, strlen(buf)+1);
	
}

void send_routes(int s)
{
	char buf[128];
	struct rspf_route *rt;
	qmark rtqm;
	
	sprintf(buf, "Routes\n");
	write(s, buf, strlen(buf)+1);
	
	sprintf(buf, "Destination    Sigbits  Cost  Iface\n");
	write(s, buf, strlen(buf)+1);
	
	rt = (struct rspf_route*)qmove_first(rt_queue, &rtqm);
	while(rt != NULL)
	{
		sprintf(buf, "%15s %7u  %4u  %5s\n", in_ntoa(rt->addr), rt->sigbits, rt->cost, rt->port);
		write(s, buf, strlen(buf)+1);
		rt = (struct rspf_route*)qmove_next(rt_queue, &rtqm);
	}
}

void send_nodegroups(int s)
{
	char buf[128];
	struct nodegroup *ng;
	qmark ngqm;
	
	sprintf(buf, "Node Groups\n");
	write(s, buf, strlen(buf)+1);
	
	sprintf(buf, "Group               Cost  Iface\n");
	write(s, buf, strlen(buf)+1);
	
	ng = (struct nodegroup*)qmove_first(nodgrpq, &ngqm);
	while(ng != NULL)
	{
		sprintf(buf, "%15s/%2u  %4u  %s\n", in_ntoa(ng->addr), ng->sigbits, ng->cost, ng->iface);
		write(s, buf, strlen(buf)+1);
		ng = (struct nodegroup*)qmove_next(nodgrpq, &ngqm);
	}
	
}